import asyncio
import argparse
import json
import sys
import os
import pkg_resources
import time
from typing import List, Tuple, Dict

from tokmon.tokmon import TokenMonitor
from tokmon.costcalculator import CostCalculator

PROG_NAME = "tokmon"

BOLD = "\033[1m"
RESET = "\033[0m"
MAGENTA = "\033[35m"
GRAY = "\033[90m"
GREEN = "\033[32m"

def bold(s:str):
    return f"{BOLD}{s}{RESET}"

def color(s:str, color:str, bold:bool = True):
    if bold:
        return f"{BOLD}{color}{s}{RESET}"
    else:
        return f"{color}{s}{RESET}"

def print_usage_report(monitored_invocation:str, cost_summary: Dict):
    models = cost_summary["models"]
    pricing = cost_summary["pricing_data"]
    total_cost = cost_summary["total_cost"]
    total_usage = cost_summary["total_usage"]

    cost_str = f"${total_cost:.6f}"
    report_header = f"{PROG_NAME} cost report:"

    print(f"""
{color(report_header, GREEN)}
{color('='*80, GRAY, bold=False)}
{bold("Monitored invocation")}: {monitored_invocation}
{bold("Models")}: {models}
{bold("Total Usage")}: {total_usage}
{bold("Pricing")}: {pricing}
{color("Total Cost", MAGENTA)}: {color(cost_str, MAGENTA)}
{color('='*80, GRAY, bold=False)}
""")

OPENAI_API_PATH = "https://api.openai.com"
DEFAULT_JSON_OUT_PATH = "/tmp"

def cli():
    """
    The {PROG_NAME} utility can be used to monitor the cost of OpenAI API calls made by a program.
    After te program has finished running, the {PROG_NAME} will print the total cost of the program.
    """
    parser = argparse.ArgumentParser(description="A utility to monitor OpenAI token cost of a target program.",
                                     add_help=False)

    current_time = int(time.time())
    default_json_out = os.path.join("/tmp", f"{PROG_NAME}_usage_summary_{current_time}.json")

    parser.add_argument("program_name", nargs="?", help="The name of the monitored program")
    parser.add_argument("args", nargs=argparse.REMAINDER, help="The command and arguments to run the monitored program")
    parser.add_argument("-p", "--pricing", type=str, help="Path to a custom OpenAI pricing JSON file", default=None)
    parser.add_argument("-v", "--verbose", action="store_true", help="Print verbose output")
    parser.add_argument("-j", "--json_out", type=str, help="Path to a JSON file to write the cost summary to. Saves to /tmp by default", default=DEFAULT_JSON_OUT_PATH)
    parser.add_argument("-n", "--no_json", action="store_true", help="Do not write a cost summary to a JSON file")
    parser.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    args = parser.parse_args()

 
    if args.json_out and args.no_json:
        parser.error("Cannot use --json_out and --no_json together")
        sys.exit(1)

    if not args.program_name:
        parser.print_help()
        sys.exit(1)

    # Note: openai-pricing data may go out of date
    if args.pricing:
        pricing_json = args.pricing
    else:
        pricing_json = pkg_resources.resource_filename(PROG_NAME, "openai-pricing.json")

    with open(pricing_json, "r") as f:
        pricing = json.load(f)

    monitored_prog = f"{args.program_name} { ' '.join(args.args) if args.args else ''}"

    # Instantiate the token monitor
    tokmon = TokenMonitor(OPENAI_API_PATH, args.program_name, *args.args, verbose=args.verbose)

    try:
        monitoring_str = f"[{PROG_NAME}] Monitoring token usage for {color(monitored_prog, GREEN)} ..."
        print(f"{color(monitoring_str, MAGENTA)}")

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        loop.run_until_complete(tokmon.start_monitoring())
    except KeyboardInterrupt:
        interrupted_str = f"\n[{PROG_NAME}] Interrupted. Generating token usage report ..."
        print(f"{color(interrupted_str, MAGENTA)}")
    finally:
        tokmon.stop_monitoring()
        
        # Get the usage summary
        usage_summary = tokmon.usage_summary()
        
        # If no usage was detected, print a message and exit
        if len(usage_summary) == 0:
            status_str = f"[{PROG_NAME}] No OpenAI API calls detected for `{monitored_prog}`."
            print(f"{color(status_str, MAGENTA)}")
            return

        # Print usage report to the terminal
        cost_summary = calculate(usage_summary, pricing)
        print_usage_report(monitored_prog, cost_summary)

        # Write usage report to a JSON file
        if args.json_out and not args.no_json:
            json_out_filename = f"{PROG_NAME}_usage_summary_{current_time}.json"
            out_dir_path = args.json_out
            if not os.path.exists(out_dir_path):
                print(f"** Path does not exist: {out_dir_path}, falling back to {DEFAULT_JSON_OUT_PATH}")
                out_dir_path = DEFAULT_JSON_OUT_PATH
            json_out_path = os.path.join(out_dir_path, f"{json_out_filename}")
            print(f"Writing cost summary to JSON file: {color(json_out_path, GREEN)} {color('(run with --no_json to disable this behavior)', GRAY)}")
            with open(json_out_path, "w") as f:
                 json.dump(cost_summary, f, indent=4)

def calculate(usage_summary: List[Tuple[Dict, Dict]], pricing: Dict):
    costCalculator = CostCalculator(pricing)
    usage_with_cost = costCalculator.calculate_cost(usage_summary)
    return usage_with_cost

if __name__ == '__main__':
    cli()


import asyncio
import os
import json
import subprocess
import typing
import time
import socket
from typing import Callable, Any, List, Tuple, Dict

import tiktoken

from mitmproxy import http, options
from mitmproxy.tools.dump import DumpMaster

def find_available_port(start_port: int):
    """
    To allow multiple instances of tokmon to run concurrently.
    """
    port = start_port
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(('localhost', port)) != 0:
                return port
            port += 1

def count_tokens_in_json(encode_fn: Callable[[str], List[str]], data: Any) -> int:
    token_count = 0
    stack = [data]

    while stack:
        current = stack.pop()

        if isinstance(current, dict):
            for key, value in current.items(): # 'key' is unused
                # empricially, the keys seem to not be counted towards the token count
                # so we are not including this `token_count += count_tokens(key)`
                stack.append(value)
        elif isinstance(current, list):
            stack.extend(current)
        elif isinstance(current, str):
            token_count += len(encode_fn(current))
        else:
            token_count += len(encode_fn( str(current) ) )

    return token_count

PORT = find_available_port(7878)

class TokenMonitor:
    def __init__(self, target_url: str, program_name: str, *args: tuple, verbose:bool = False):
        self.mitm = None
        self.target_url = target_url
        self.program_name = program_name
        self.args = args
        self.process = None
        self.using_stream = False
        self.verbose = verbose
        self.history: List[Tuple[Dict, Dict]] = []
        self.current_request = None

    # For handling streaming requests:
    # def responseheaders(self, flow: http.HTTPFlow):
    #     if self.target_url in flow.request.pretty_url:
    #         content_type = flow.response.headers.get("Content-Type", "")
    #         if "text/event-stream" in content_type:
    #             flow.response.stream = True

    def request(self, flow: http.HTTPFlow):
        self.handle_request(flow)

    def response(self, flow: http.HTTPFlow):
        self.handle_response(flow)
            
    def handle_request(self, flow: http.HTTPFlow):
        if self.target_url in flow.request.pretty_url:
            try:
                if self.current_request is not None:
                    print("Warning: multiple requests in flight. This is not supported.")
                
                request_data = json.loads(flow.request.content)
                self.current_request = request_data

                if self.verbose:
                    print(request_data)
                
                self.using_stream = request_data["stream"]

            except json.JSONDecodeError:
                print("Failed to parse request data as JSON")

    def handle_response(self, flow: http.HTTPFlow):
        if not flow.request.url.startswith(self.target_url):
            return
    
        model = ""
        content = ""
        usage = None

        if flow.response.text:
            if self.using_stream:
                model, content, usage = self.handle_stream_response(flow.response.text)
            else:    
                response_data = json.loads(flow.response.text)
                model = response_data["model"]
                content = response_data["choices"][0]["message"]["content"]
                usage = response_data["usage"]
        else:
            raise Exception("No response data")
        
        request = self.current_request
        response = {
            "model": model,
            "messages": [{"role": "assistant", "content": content}],
            "usage": usage
        }

        self.history.append((request, response))
        self.current_request = None

        if self.verbose:
            print(response)

    async def run_monitored_program(self) -> bool:
        env = os.environ.copy()

        # mitproxy automatically generates a CA cert and stores it in ~/.mitmproxy ...
        mitmproxy_path = os.path.expanduser("~/.mitmproxy")
        mitmproxy_abs_path = os.path.abspath(mitmproxy_path)
        # ... but this coroutine might be called before mitmproxy has had a chance to generate the cert ... ¯\_(ツ)_/¯.
        timeout_seconds = 2
        start = time.time()
        while not os.path.exists(mitmproxy_abs_path):
            await asyncio.sleep(0.1) # yield to other coroutines
            if time.time() - start > timeout_seconds:
                raise Exception("\n\n*** Error: Cert file wasn't created in time. Try to run `mitmproxy` manually first, and then try running tokmon again.\n\n")

        ca_cert_file = "mitmproxy-ca-cert.pem"
        ca_cert_abs_path = os.path.join(mitmproxy_abs_path, ca_cert_file)

        # set the HTTP proxy environment variables
        env["HTTP_PROXY"] = f"http://localhost:{PORT}"
        env["HTTPS_PROXY"] = f"http://localhost:{PORT}"

        # add `mitmproxy`'s CA cert to the environment variables of the monitored program
        env["REQUESTS_CA_BUNDLE"] = ca_cert_abs_path # for monitored programs using Python's Requests Library
        env["NODE_EXTRA_CA_CERTS"] = ca_cert_abs_path # for monitored programs using Node.js

        if self.program_name == 'curl':
            print()
            args = ['--cacert', ca_cert_abs_path] + [arg for arg in self.args]
        else:
            args = [arg for arg in self.args]

        try:
            self.process = subprocess.Popen([self.program_name] + args, env=env)
            return True
        except FileNotFoundError:
            print(f"[tokmon] Error: Program not found '{self.program_name}'. Did you type its path and name correctly?")
        except Exception as e:
            print(e)
            
        return False
    
    def encode(self, model, text):
        """
        Learn more: https://github.com/openai/tiktoken
        """
        tokenizer = tiktoken.encoding_for_model(model)
        return tokenizer.encode(text)

    def handle_stream_response(self, raw_messages: typing.List[str]):
        """
        When streaming, OpenAI's API doesn't return usage data.
        To work around this, we use tiktoken directly.

        See: https://community.openai.com/t/usage-info-in-api-responses/18862/11
        """
        
        assert self.current_request is not None

        model = None

        completion_tokens = 0
        completion_content = ""

        # mitmproxy buffers the returned SSE chunks as one big string
        for msg in raw_messages.split("\n"):
            if msg.startswith("data:"):
                try:
                    msg = msg[5:]
                    if "".join(msg.split()) == "[DONE]":
                        break
                    msg = json.loads(msg)
                    model = msg["model"]
                    
                    choice = msg["choices"][0]
                    if "delta" in choice:
                        if "content" in choice["delta"]:
                            tokens = choice["delta"]["content"]
                            completion_content += tokens
                            encoded_tokens = self.encode(model, tokens)
                            completion_tokens += len(encoded_tokens)

                except json.JSONDecodeError as e:
                    print(f"\n\n ! ! Failed to parse response data as JSON {e} --- <{msg}> ! !\n\n")
                except Exception as e:
                    print(f"\n\n ! ! Failed to parse response data: {e} ! !\n\n")
                    raise e
        
        encode_lambda = lambda text: self.encode(model, text)
        prompt_tokens = count_tokens_in_json(encode_lambda, self.current_request)
        total_tokens = prompt_tokens + completion_tokens
        
        # mimic the usage data returned by the API in the non streaming case
        usage = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens
        }
        return model, completion_content, usage
    
    async def start_monitoring(self):        
        opts = options.Options(listen_host='0.0.0.0', listen_port=PORT)
        self.mitm = DumpMaster(opts, with_termlog=False, with_dumper=False)
        self.mitm.addons.add(self)

        async def run_mitmproxy():
            try:
                await self.mitm.run()
            except KeyboardInterrupt:
                pass
            except Exception as e:
                print(f"Exception while running mitmproxy: {e}")
            finally:
                self.stop_monitoring()

        async def wait_subprocess():
            success = await self.run_monitored_program()
            if success:   
                while self.process.poll() is None:
                    await asyncio.sleep(1)
            self.stop_monitoring()

        await asyncio.gather(run_mitmproxy(), wait_subprocess())
    
    def stop_monitoring(self):
        if self.process:
            self.process.terminate()
        self.mitm.shutdown()

    def usage_summary(self) -> List[Tuple[Dict, Dict]]:
        return self.history



from typing import List, Tuple, Dict

class CostCalculator:
    def __init__(self, pricing_data: Dict[str, Dict[str, float]]):
        self.pricing_data = pricing_data

    def calculate_cost_for_tokens(self, tokens, price, per_tokens):
        return (float(tokens) / per_tokens) * price

    def calculate_round_trip_cost(self, request: Dict, response: Dict):
        """
        Calculate cost & usage for a single round trip (request -> response)
        """
        model = response["model"]
        usage_data = response["usage"]
        
        prompt_tokens = usage_data["prompt_tokens"]
        completion_tokens = usage_data["completion_tokens"]
        total_tokens = usage_data["total_tokens"]
        assert total_tokens == prompt_tokens + completion_tokens, "Total tokens does not match prompt + completion tokens"

        # prepare pricing data
        model_pricing_data = self.pricing_data[model]
        per_tokens = model_pricing_data["per_tokens"]

        if "prompt_cost" in model_pricing_data:
            """Model differentiates between prompt and completion costs"""
            prompt_price = model_pricing_data["prompt_cost"]
            completion_price = model_pricing_data["completion_cost"]

            prompt_cost = self.calculate_cost_for_tokens(prompt_tokens, prompt_price, per_tokens)
            completion_cost = self.calculate_cost_for_tokens(completion_tokens, completion_price, per_tokens)
            total_cost = prompt_cost + completion_cost
        else:
            """Model has a single cost for all tokens"""
            price = model_pricing_data["cost"]
            total_cost = self.calculate_cost_for_tokens(total_tokens, price, per_tokens)

        messages = request["messages"] + response["messages"]

        return model_pricing_data, {
            "model": model,
            "usage": {"prompt_tokens": prompt_tokens, "completion_tokens": completion_tokens, "total_tokens": total_tokens },
            "cost": total_cost,
            "messages": messages
        }

    def calculate_cost(self, usage_data:  List[Tuple[Dict, Dict]]):
        """
        Calculate cost & usage for all of (request, response) pairs, return a summary
        """
        total_cost = 0.0
        total_prompt_tokens = 0
        total_completion_tokens = 0
        total_tokens = 0
        pricing_data = {}
        models = set()
        raw_data = []

        for request, response in usage_data:
            model_pricing, round_trip_cost = self.calculate_round_trip_cost(request, response)
            usage = round_trip_cost["usage"]
            total_prompt_tokens += usage["prompt_tokens"]
            total_completion_tokens += usage["completion_tokens"]
            total_tokens += usage["total_tokens"]
            total_cost += round_trip_cost["cost"]
            model = round_trip_cost["model"]
            models.add(model)
            pricing_data[model] = model_pricing
            raw_data.append(round_trip_cost)

        return {
            "total_cost": total_cost,
            "total_usage": {
                "total_prompt_tokens": total_prompt_tokens,
                "total_completion_tokens": total_completion_tokens,
                "total_tokens": total_tokens,
            },
            "pricing_data": str(pricing_data),
            "models": list(models),
            "raw_data": raw_data
        }


from setuptools import setup, find_packages
import os

# Read the contents of the README.md file
with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='tokmon',
    version='0.1.4',
    packages=find_packages(),
    install_requires=[
        'mitmproxy',
        'tiktoken',
    ],
    package_data={
        "tokmon": ["openai-pricing.json"],
    },
    entry_points={
        'console_scripts': [
            "tokmon = tokmon.cli:cli",
        ],
    },
    description = "tokmon is a CLI utility to monitor OpenAI token usage and costs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yagil/tokmon",
        project_urls={  # Optional
        "Bug Reports": "https://github.com/yagil/tokmon/issues",
        "Source": "https://github.com/yagil/tokmon/",
    },
)